module MUD (uniformPDF, gaussianPDF, kde, wme, getPF, updated) where

-- Function to evaluate Uniform PDF at a point x with range [a, b]
uniformPDF :: Double -> Double -> Double -> Double
uniformPDF a b x
  | x >= a && x <= b = 1 / (b - a)
  | otherwise        = 0

-- Function to evaluate Gaussian PDF at a point x with mean mu and standard deviation sigma
gaussianPDF :: Double -> Double -> Double -> Double
gaussianPDF x mu sigma = (1 / sqrt (2 * pi * sigma**2)) * exp (-(x - mu)**2 / (2 * sigma**2))

-- Kernel Density Estimation function
kde :: [Double] -> Double -> (Double -> Double)
kde ctlpoints bandwidth = \x -> (1 / (n * actualBandwidth)) * sum [gaussianPDF x xi actualBandwidth | xi <- ctlpoints]
  where
    n = fromIntegral (length ctlpoints)
    actualBandwidth
      | bandwidth == 0 = 1.06 * stdDev ctlpoints * n ** (-1/5)
      | otherwise      = bandwidth

-- Calculate standard deviation
stdDev :: [Double] -> Double
stdDev xs = sqrt $ sum [(x - mean) ** 2 | x <- xs] / n
  where
    n = fromIntegral (length xs)
    mean = sum xs / n

-- Estimate the push-forward distribution
getPF :: (Double -> Double) -> [Double] -> (Double -> Double)
getPF f x = kde y 0
  where
    y = map f x

-- Weighted Mean Error Functional
wme :: Double -> [Double] -> Double -> Double
wme y d s = (1 / sqrt n) * sum [ (y - di) / std | di <- d ]
  where
    n = fromIntegral (length d)
    std
      | s == 0    = stdDev d
      | otherwise = s

-- Updated density computation.
updated :: (Double -> Double) -> (Double -> Double) -> [Double] -> [Double]
updated f d x = [ i * (o / p) | (i, o, p) <- zip3 a b c]
  where
    y = map f x
    a = map (\xi -> gaussianPDF xi 0 1) x
    b = map (\yi -> gaussianPDF yi 0 1) y
    c = map d y
